/*
 * usbcore.c : USB protocol handler
 *   by uratan! 2008.9.18
 */
 	/*
	 * I WILL SUPPORT ONLY:
	 *   single Configuration Descriptor, and
	 *   no Alternate Settings
	 */
#include "common.h"
#include "dscr.h.S"
#include "usb.h"

/*
 * my local debug flag
 */
#define myUDEBUG  (1 && gUDEBUG)
#define RPRINTF    if(myUDEBUG)rprintf

/*
 * EPaddr for Control-Pipe
 */
#define EP_CTRL_OUT  (0x00 | 0)
#define EP_CTRL_IN   (0x80 | 0)

/*
 * functions
 */
void ucore_InitialCore(void);
void ucore_ResetCore(void);
void ucore_SuspendCore(void);
void ucore_TryPowerDown(void);
void ucore_SetupCommand(void);
void ucore_ProceedSetupCommand(void);
void ucore_start_DataIN(u_char *buff, u_long len);
int ucore_start_DataOUT(u_char *buff, u_long len);
void ucore_start_StatusIN(void);
void ucore_STALL_EP0(void);
static int ucore_GetStatus(void);
static int ucore_SetClrFeature(u_long sc);
static int ucore_GetDescriptor(void);
static u_char *ucore_Get_NthStringDscr(u_char index);
static int ucore_SetConfiguration(void);
static int ucore_SetInterface(void);
void ucore_isr_EP0_in(u_long unused);
void ucore_isr_EP0_out(u_long unused);
static void ucore_set_IN_data(void);
static int ucore_feed_OUT_data(void);

/*
 * variables
 */
u_char *USB_DeviceDescriptor;
u_char *USB_ConfigDescriptor;
u_char *USB_StringDescriptor;
u_char **USB_list_of_EpDscrs;
u_short USB_DeviceStatus;
u_char  USB_Configuration;	/* 0 or configuration number */
u_char  USB_NumInterfaces;	/* of selected configuration */
u_char  USB_AltSetting;		/* dummy */

/*
 * make state machine for Control-Transfer
 */
#define CTS_IDLE     0		/* only SETUP will be accepted */
#define CTS_DATAIN   1		/* expect IN from host */
#define CTS_DATAOUT  2		/* expect OUT from host */
#define CTS_STATOUT  3		/* expect OUT from host */
#define CTS_STATIN   4		/* expect IN from host */
static int ct_state;

static u_long ct_len_requested;
static int ct_need_zero_packet;
static u_char *ct_data;
static u_long ct_residue;

/*
 * data buffers for Control-Transfer
 */
u_char SETUPDAT[8];
u_char EP0BUF[DSCR_MPSIZE_EP0];

/*
 * externs
 */
/* in dscr.S */
extern u_long *DeviceDscr, *ConfigDscr, *StringDscr;
/* in main.c */
extern VOLATILE int GotCOD;		/* Control-OUT Data */
/* in pdown.c */
extern int power_down(int en_eint);
/* in usbhw.c */
extern void uhw_WakeUp(void);
extern void uhw_SetAddress(u_char addr);
extern void uhw_Configure(int cfg);
extern void uhw_EnableEP(u_char EPaddr, u_long max_packet_size);
extern u_long uhw_IsEnabledEP(u_char EPaddr);
extern u_long uhw_IsStallEP(u_char EPaddr);
extern void uhw_SetStallEP(u_char EPaddr);
extern void uhw_ClrStallEP(u_char EPaddr);
extern void uhw_WriteEP(u_char EPaddr, u_char *pData, u_long len);
extern void uhw_arm_WriteEP(u_char EPaddr);
extern u_long uhw_ReadEP(u_char EPaddr, u_char *pData, int capacity);
extern void uhw_ReadEP_finish(u_char EPaddr);
/* in periph.c */
extern int DR_SetupCommand(void);
extern void DR_ProceedSetupCommand(void);
extern int DR_GetDescriptor(void);
///* in vcomuser.c */
//extern void vcom_USB_Configure_Event(void);
//extern void vcom_USB_Interface_Event(void);
//extern void vcom_USB_Feature_Event(void);


/*
 * initial valiables about USB core
 */
void ucore_InitialCore(void)
{
	USB_DeviceDescriptor = (u_char *)&DeviceDscr;
	USB_ConfigDescriptor = (u_char *)&ConfigDscr;
	USB_StringDescriptor = (u_char *)&StringDscr;
	USB_list_of_EpDscrs = NULL;

	ucore_ResetCore();
}

/*
 * initial USB core by USB reset
 */
void ucore_ResetCore(void)
{
	/*
	 * by USB reset, initial these 
	 */
	if(USB_ConfigDescriptor[/*bmAttributes*/ 7] & UC_SELF_POWERED)
		USB_DeviceStatus = UDS_SELF_POWERED;
	else
		USB_DeviceStatus = 0x00;
			/*
			 * reset value of UDS_REMOTE_WAKEUP is FALSE
			 */
	USB_Configuration = 0;

	ct_state = CTS_IDLE;
}

/*
 *
 */
void ucore_SuspendCore(void)
{
	ct_state = CTS_IDLE;		/* make it sure */
}

/*
 * try to enter power down mode
 */
void ucore_TryPowerDown(void)
{
	int en_eint, wakeup_by_eint;

RPRINTF("TryPowerDown()\r\n");
	/*
	 * in case of UDS_REMOTE_WAKEUP:
	 *    exit power-down mode by USB Resume or EINT2,
	 *    wake-up the host if woken by EINT2
	 * in case of !UDS_REMOTE_WAKEUP:
	 *    exit power-down mode by only USB Resume,
	 *    no need to wake-up the host
	 */
	en_eint = (USB_DeviceStatus & UDS_REMOTE_WAKEUP) ? 1 : 0;
	wakeup_by_eint = power_down(en_eint);
	if(wakeup_by_eint) {
		uhw_WakeUp();
	}
RPRINTF("wakeup\r\n");
}

/*
 * device request parser
 */
void ucore_SetupCommand(void)
{
	/*
	 * take SETUP data out from hardware buffer
	 */
	uhw_ReadEP(EP_CTRL_OUT, SETUPDAT, 8);
//	uhw_ReadEP_finish(EP_CTRL_OUT);
		/*
		 * I will release the buffer if needed in the following process.
		 * the host receives NAK by next OUT packet and wait until
		 * we are ready(==release-buffer).
		 */
	/*
	 * sample some params in the future
	 */
	ct_len_requested = SETUPDAT[6] + (SETUPDAT[7] << 8);
	ct_state = CTS_IDLE;

#if myUDEBUG
{
	int k;

	rprintf("SETUP[]:");
	for(k=0; k<8; k++) {
		rprintf(" 0x%02x", SETUPDAT[k]);
	}
	rprintf("\r\n");
}
#endif /* myUDEBUG */

	/*
	 * parse the request
	 * at first, parse device-dependant requests here
	 */
  if(DR_SetupCommand() != DONE) {
	switch(SETUPDAT[0] & UT_TYPE_MASK) {
	case UT_STANDARD:
			switch(SETUPDAT[1]) {
			case UR_GET_STATUS:
RPRINTF("GET_STAT\r\n");
					if(ucore_GetStatus() != DONE) {
						goto stall_i;
					}
					break;

			case UR_CLEAR_FEATURE:
RPRINTF("CLR_FEAT\r\n");
					if(ucore_SetClrFeature(0) != DONE) {
						goto stall_i;
					}
//					vcom_USB_Feature_Event();
					break;

			case UR_SET_FEATURE:
RPRINTF("SET_FEAT\r\n");
					if(ucore_SetClrFeature(1) != DONE) {
						goto stall_i;
					}
//					vcom_USB_Feature_Event();
					break;

			case UR_SET_ADDRESS:
RPRINTF("SET_ADDR\r\n");
					switch(SETUPDAT[0] & UT_RECIP_MASK) {
					case UT_TO_DEVICE:
							uhw_SetAddress(SETUPDAT[2]);
							ucore_start_StatusIN();
							break;
					default:
							goto stall_i;
					}
					break;

			case UR_GET_DESCRIPTOR:
RPRINTF("GET_DESC\r\n");
				if(DR_GetDescriptor() != DONE)
					if(ucore_GetDescriptor() != DONE) {
						goto stall_i;
					}
					break;

			case UR_SET_DESCRIPTOR:
RPRINTF("SET_DESC\r\n");
					goto stall_i;
					break;

			case UR_GET_CONFIGURATION:
RPRINTF("GET_CONF\r\n");
					switch(SETUPDAT[0] & UT_RECIP_MASK) {
					case UT_TO_DEVICE:
							EP0BUF[0] = USB_Configuration;
							ucore_start_DataIN(EP0BUF, 1);
							break;
					default:
							goto stall_i;
					}
					break;

			case UR_SET_CONFIGURATION:
RPRINTF("SET_CONF\r\n");
					switch(SETUPDAT[0] & UT_RECIP_MASK) {
					case UT_TO_DEVICE:
							if(ucore_SetConfiguration() != DONE) {
								goto stall_i;
							}
//							vcom_USB_Configure_Event();
							break;
					default:
							goto stall_i;
					}
					break;

			case UR_GET_INTERFACE:
RPRINTF("GET_INTF\r\n");
					switch(SETUPDAT[0] & UT_RECIP_MASK) {
					case UT_TO_INTERFACE:
							if(
								(USB_Configuration != 0) &&
								(SETUPDAT[4] < USB_NumInterfaces)
							) {
								EP0BUF[0] = USB_AltSetting;
								ucore_start_DataIN(EP0BUF, 1);
							} else {
								goto stall_i;
							}
							break;
					default:
							goto stall_i;
					}
					break;

			case UR_SET_INTERFACE:
RPRINTF("SET_INTF\r\n");
					switch(SETUPDAT[0] & UT_RECIP_MASK) {
					case UT_TO_INTERFACE:
							if(ucore_SetInterface() != DONE) {
								goto stall_i;
							}
//							vcom_USB_Interface_Event();
							break;
					default:
							goto stall_i;
					}
					break;

			default:
					goto stall_i;
			}
			break;

	case UT_CLASS:
	case UT_VENDOR:
	default:
stall_i:
			ucore_STALL_EP0();
			break;
	}
  }
RPRINTF("@@@\r\n");
}

/*
 * continue to proceed setup command after Control-OUT data received
 */
void ucore_ProceedSetupCommand(void)
{
//	switch(my_job_stat) {
//	case XXX:
//			...
//			break;
//	case YYY:
//			...
//			break;
//	default:
			DR_ProceedSetupCommand();
//			break;
//	}
RPRINTF("###\r\n");
}

/*
 * initiate DataIN
 */
void ucore_start_DataIN(u_char *buff, u_long len)
{
	/*
	 * round data length by length in SETUP data
	 */
	if(len > ct_len_requested) {
		len = ct_len_requested;
	}

	/*
	 * setup Control-Transfer params
	 */
	ct_data = buff;
	ct_residue = len;

	/*
	 * add zero-length-packet if the last data length is just fit
	 * to packet size while the host expect more.
	 */
	if(
		(len != ct_len_requested) &&
		((len % DSCR_MPSIZE_EP0) == 0) &&
		(len != 0)
	) {
		ct_need_zero_packet = TRUE;
	} else {
		ct_need_zero_packet = FALSE;
	}

	/*
	 * I expect IN from the host
	 */
	ct_state = CTS_DATAIN;

	/*
	 * initiate first DataIN packet
	 */
	ucore_set_IN_data();
}

/*
 * initiate DataOUT
 */
int ucore_start_DataOUT(u_char *buff, u_long len)
{
	/*
	 * check data length
	 */
	if(len != ct_len_requested) {
RPRINTF("wrong len\r\n");
		return(-1);
	}

	/*
	 * setup Control-Transfer params
	 */
	ct_data = buff;
	ct_residue = len;

	/*
	 * I expect OUT from the host
	 */
	ct_state = CTS_DATAOUT;

	/*
	 * parse the request finished, make OUT-buffer free
	 * to receive next DataOUT
	 */
	uhw_ReadEP_finish(EP_CTRL_OUT);

	return(0);
}

/*
 * initiate status IN
 */
void ucore_start_StatusIN(void)
{
	/*
	 * I expect IN from the host
	 */
	ct_state = CTS_STATIN;

	/*
	 * initiate status-in packet
	 */
	uhw_WriteEP(EP_CTRL_IN, NULL, 0);
	uhw_arm_WriteEP(EP_CTRL_IN);
}

/*
 * stall both Endpoint 0 in/out
 */
void ucore_STALL_EP0(void)
{
	uhw_SetStallEP(EP_CTRL_IN);		/* for IN */
	uhw_SetStallEP(EP_CTRL_OUT);	/* for OUT */
	ct_state = CTS_IDLE;
RPRINTF("STALL_EP0()\r\n");
		/*
		 * the stall bits are automatically cleared when next SETUP received
		 */
}

/*
 * parse Get Status request
 */
static int ucore_GetStatus(void)
{
	u_long n;

	switch(SETUPDAT[0] & UT_RECIP_MASK) {
	case UT_TO_DEVICE:
			EP0BUF[0] = (USB_DeviceStatus>>0) & 0xff;
			EP0BUF[1] = (USB_DeviceStatus>>8) & 0xff;
			break;

	case UT_TO_INTERFACE:
			if((USB_Configuration != 0) && (SETUPDAT[4] < USB_NumInterfaces)) {
				EP0BUF[0] = 0;
				EP0BUF[1] = 0;
			} else {
				return(YET);
			}
			break;

	case UT_TO_ENDPOINT:
			n = SETUPDAT[4] & 0x8F;
			if(
				((USB_Configuration != 0) || ((n & 0x0F) == 0)) &&
				uhw_IsEnabledEP(n)
			) {
				EP0BUF[0] = uhw_IsStallEP(n);
				EP0BUF[1] = 0;
			} else {
				return(YET);
			}
			break;

	default:
			return(YET);
	}

	ucore_start_DataIN(EP0BUF, 2);
	return(DONE);
}

/*
 * parse Set/Clear Feature request
 *   pass 0 for clear, 1 for set
 */
static int ucore_SetClrFeature(u_long sc)
{
	u_long n;

	switch(SETUPDAT[0] & UT_RECIP_MASK) {
	case UT_TO_DEVICE:
			if((SETUPDAT[2]+(SETUPDAT[3]<<8)) == UF_REMOTE_WAKEUP) {
				if(sc) {
					USB_DeviceStatus |= UDS_REMOTE_WAKEUP;
				} else {
					USB_DeviceStatus &= ~UDS_REMOTE_WAKEUP;
				}
			} else {
				return(YET);
			}
			break;

	case UT_TO_INTERFACE:
			return(YET);

	case UT_TO_ENDPOINT:
			n = SETUPDAT[4] & 0x8F;
			if(
				(USB_Configuration != 0) &&
				((n & 0x0F) != 0) &&
				uhw_IsEnabledEP(n)
			) {
				if((SETUPDAT[2]+(SETUPDAT[3]<<8)) == UF_ENDPOINT_HALT) {
					if(sc) {
						uhw_SetStallEP(n);
					} else {
						uhw_ClrStallEP(n);
					}
				} else {
					return(YET);
				}
			} else {
				return(YET);
			}
			break;

	default:
			return(YET);
	}

	ucore_start_StatusIN();
	return(DONE);
}

/*
 * parse Get Descriptor request
 */
static int ucore_GetDescriptor(void)
{
	u_char *dscr;
	u_long len;

	switch(SETUPDAT[0] & UT_RECIP_MASK) {
	case UT_TO_DEVICE:
			switch(SETUPDAT[3]) {
			case UDESC_DEVICE:
RPRINTF("UDESC_DEVICE\r\n");
					dscr = USB_DeviceDescriptor;
					len = dscr[/*bLength*/ 0];
					break;

			case UDESC_CONFIG:
RPRINTF("UDESC_CONFIG\r\n");
					if(SETUPDAT[2] != 0) {
						return(YET);
					}
					dscr = USB_ConfigDescriptor;
					len = dscr[2] + (dscr[3]<<8);	/* wTotalLength */
					break;

			case UDESC_STRING:
RPRINTF("UDESC_STRING\r\n");
					dscr = ucore_Get_NthStringDscr(SETUPDAT[2]);
					if(dscr == NULL) {
						return(YET);
					}
					len = dscr[/*bLength*/ 0];		/* bLength */
					break;

			default:
					return(YET);
			}
			break;

	case UT_TO_INTERFACE:
	default:
			return(YET);
	}

	ucore_start_DataIN(dscr, len);
	return(DONE);
}

/*
 * search Nth string descriptor
 */
static u_char *ucore_Get_NthStringDscr(u_char index)
{
	u_char *dscr;

	dscr = USB_StringDescriptor;

	while(dscr[1 /*type*/] == UDESC_STRING) {
		if(index == 0)
			return(dscr);

		index--;
		dscr += dscr[0 /*length*/];
	}

	return(NULL);	/* not found */
}

/*
 * parse Set Configuration request
 */
static int ucore_SetConfiguration(void)
{
	u_char **ep_list, *dscr;
	u_long n;

	/*
	 * only configuration one is permitted by this firmware
	 */
	if(SETUPDAT[2] == 1) {
		/*
		 *
		 */
		USB_Configuration = SETUPDAT[2];
		USB_NumInterfaces = USB_ConfigDescriptor[/*bNumInterfaces*/ 4];
		USB_AltSetting = 0;

		uhw_Configure(TRUE);

		/*
		 *
		 */
		if(USB_list_of_EpDscrs != NULL) {
			ep_list = USB_list_of_EpDscrs;
			while(*ep_list != NULL) {
				/*
				 * enable Endpoints about default setting
				 */
				dscr = *ep_list;
				n = dscr[/*bEndpointAddress*/ 2] & 0x8F;
				uhw_EnableEP(n, dscr[4] + (dscr[5]<<8));	/* wMaxPacketSize */
				uhw_ClrStallEP(n);

				ep_list++;
			}
		}

		ucore_start_StatusIN();
		return(DONE);
	}

	if(SETUPDAT[2] == 0) {
		/*
		 * configuration canceled
		 */
		USB_Configuration = 0;

		uhw_Configure(FALSE);

		ucore_start_StatusIN();
		return(DONE);
	}

	return(YET);
}

/*
 * parse Set Interface request
 */
static int ucore_SetInterface(void)
{
	u_char **ep_list, *dscr;
	u_long n;

	/*
	 * am I configured ?
	 */
	if(USB_Configuration == 0) {
		return(YET);
	}

	/*
	 * only setting zero is permitted by this firmware
	 */
	if(SETUPDAT[2] != 0) {
		return(YET);
	}
	USB_AltSetting = SETUPDAT[2];

	/*
	 * clear all EndpointHalts here
	 */
	if(USB_list_of_EpDscrs != NULL) {
		ep_list = USB_list_of_EpDscrs;
		while(*ep_list != NULL) {
			/*
			 * enable Endpoints about default setting
			 */
			dscr = *ep_list;
			n = dscr[/*bEndpointAddress*/ 2] & 0x8F;
			uhw_ClrStallEP(n);

			ep_list++;
		}
	}

	ucore_start_StatusIN();
	return(DONE);
}

/* =============================================================== */

/*
 * Endpoint 0 in interrupt handler
 */
void ucore_isr_EP0_in(u_long unused)
{
RPRINTF("!0i\r\n");

	switch(ct_state) {
	case CTS_DATAIN:
			/*
			 * specified DataIN single packet has finished
			 */
			if((ct_residue != 0) || ct_need_zero_packet) {
				/*
				 * continue to send data
				 */
				ucore_set_IN_data();
			} else {
				/*
				 * all data requeted has sent, move the state
				 */
				ct_state = CTS_STATOUT;

				/*
				 * permit to receive status-OUT packet
				 */
				uhw_ReadEP_finish(EP_CTRL_OUT);
			}
			break;

	case CTS_STATIN:
			/*
			 * do nothing about 'status-IN-has-done'
			 */
			ct_state = CTS_IDLE;
			break;

	default:
			/*
			 * unexpected token has come, only ignore...
			 */
RPRINTF("XXX\r\n");
			ct_state = CTS_IDLE;
			break;
	}
}

/*
 * Endpoint 0 out interrupt handler
 */
void ucore_isr_EP0_out(u_long unused)
{
RPRINTF("!0o\r\n");

	switch(ct_state) {
	case CTS_DATAOUT:
			/*
			 * expected DataOUT packet has arrived, try feed the data
			 */
			if(ucore_feed_OUT_data() != 0) {
				goto error;
			}

			if(ct_residue == 0) {
				/*
				 * all data requested has received,
				 * continue to parse SETUP command
				 */
				ct_state = CTS_IDLE;
				GotCOD = TRUE;
			} else {
				/*
				 * permit to receive next OUT packet
				 */
				uhw_ReadEP_finish(EP_CTRL_OUT);
			}
			break;

	case CTS_STATOUT:
			/*
			 * always ignore 'status-OUT-has-come'
			 *   (following next SETUP shall be always
			 *    'Packet over-written', don't mind)
			 */
			ct_state = CTS_IDLE;
			break;

	default:
			/*
			 * unexpected token has come, only ignore...
			 */
error:
RPRINTF("XXXXX\r\n");
			ct_state = CTS_IDLE;
			break;
	}
}

/*
 * prepare return data for next DataIN
 */
	/*
	 * --- USB Specification, Revision 1.1, Chap. 5.5.3
	 * The Data stage of a control transfer from an Endpoint to the
	 * host is complete when the Endpoint does one of the following:
	 *   1) Has transferred exactly the amount of data specified
	 *      during the Setup stage
	 *   2) Transfers a packet with a payload size less than
	 *      wMaxPacketSize or transfers a zero-length packet.
	 */
static void ucore_set_IN_data(void)
{
	u_long len;

	/*
	 * calc data length for this packet
	 */
	if(ct_residue > DSCR_MPSIZE_EP0) {
		len = DSCR_MPSIZE_EP0;
	} else {
		len = ct_residue;
	}
RPRINTF("DataIn:[%d] %02x %02x..\r\n", len, ct_data[0], ct_data[1]);

	/*
	 * write the data into the buffer
	 */
	uhw_WriteEP(EP_CTRL_IN, ct_data, len);

	/*
	 * how many data do I send now
	 */
	if(ct_residue == 0) {
		/*
		 * zero-length-packet will be sent just now
		 */
		ct_need_zero_packet = FALSE;
	} else {
		/*
		 * advance pointers for next operation
		 */
		ct_data += len;
		ct_residue -= len;
	}

	/*
	 * at last, arm the Endpoint
	 *   (to prevent 'transfer-finished' interrupt among above process)
	 */
	uhw_arm_WriteEP(EP_CTRL_IN);
}

/*
 * feed data received by DataOUT
 */
	/*
	 * --- USB Specification, Revision 1.1, Chap. 5.5.3
	 * The Data stage of a control transfer from the host to an
	 * Endpoint is complete when all of the data has been transferred.
	 * If the Endpoint receives a larger-than-expected data payload
	 * from the host, it halts the pipe.
	 */
static int ucore_feed_OUT_data(void)
{
	u_long len;

	/*
	 * take received data out
	 */
	len = uhw_ReadEP(EP_CTRL_OUT, ct_data, ct_residue);
RPRINTF("DataOut:[%d] %02x %02x..\r\n", len, ct_data[0], ct_data[1]);
		/*
		 * above uhw_ReadEP(,,9bytes) takes 14.0usec (24MHz+24MHz)
		 * where, xmemcpy(,,8bytes) takes 5.2usec in periph.c
		 *   accessing USB hardware buffers may be slow, but not so slow...
		 */

	/*
	 * does the host send bigger data than expected ?
	 */
	if(len > ct_residue) {
		return(-1);
	}

	/*
	 * advance pointers for next operation
	 */
	ct_data += len;
	ct_residue -= len;

	return(0);
}
